/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1994 - 2000, 2001 by Ralf Baechle
 * Copyright (C) 1999, 2000 Silicon Graphics, Inc.
 * Copyright (C) 2001 MIPS Technologies, Inc.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/sys.h>

#include <asm/addrspace.h>
#include <asm/asm.h>
#include <asm/cacheops.h>
#include <asm/current.h>
#include <asm/errno.h>
#include <asm/mipsregs.h>
#include <asm/page.h>
#include <asm/pgtable-bits.h>
#include <asm/stackframe.h>
#include <asm/processor.h>
#ifdef CONFIG_KBUILD_2_5
#include "asm-offsets.h"
#endif
#include <asm/regdef.h>
#include <asm/fpregdef.h>
#include <asm/unistd.h>
#include <asm/isadep.h>
#include <asm/war.h>


		.text
		.align 4
		.set	push
		.set	reorder
FEXPORT(ret_from_irq)
#ifdef CONFIG_CISR_STACK
		DECR_ISR_CNTR
		lw	t0, PT_STATUS(sp)	# returning to kernel mode?
		andi	t0, t0, KU_USER
		beqz	t0, restore_all
			
		# copy PT_SIZE contents from $sp (isr_stack), to
		#  to thread's kernel stack
		# and change $sp to thread's kernel stack
		lw	v1, kernelsp		# v1, top of stack
		subu	v0, v1, PT_SIZE		# v0, bottom of stack
		move	t0, sp

1:		subu	t2, v1, v0
		bltz	t2, 2f
		lw	t1, 0(t0)
		sw	t1, 0(v0)
		addu	t0, t0, 4
		addu	v0, v0, 4
		b	1b
2:		
		lw	sp, kernelsp
		subu	sp, sp, PT_SIZE
#endif
FEXPORT(ret_from_exception)
		lw	t0, PT_STATUS(sp)	# returning to kernel mode?
		andi	t0, t0, KU_USER
		beqz	t0, restore_all

FEXPORT(ret_from_sys_call)		# here to prevent code duplication
#ifdef CONFIG_KSTACK_GUARD
                #to see if there are any kstack overflows
                jal     kstack_guard_check 
#endif
ret_from_schedule:
		mfc0	t0, CP0_STATUS	# need_resched and signals atomic test
		ori	t0, t0, 1
		xori	t0, t0, 1
		mtc0	t0, CP0_STATUS
		SSNOP; SSNOP; SSNOP

		lw	v0, TASK_NEED_RESCHED($28)
		lw	v1, TASK_SIGPENDING($28)
		bnez	v0, reschedule
		bnez	v1, signal_return
restore_all:	.set	noat
		RESTORE_ALL_AND_RET
		.set	at

/* Put this behind restore_all for the sake of the branch prediction.  */
signal_return:
		.type	signal_return, @function

		mfc0	t0, CP0_STATUS
		ori	t0, t0, 1
		mtc0	t0, CP0_STATUS

		move	a0, zero
		move	a1, sp
		jal	do_signal
		b	restore_all

reschedule:
		jal	schedule
		b	ret_from_schedule

/*
 * Common spurious interrupt handler.
 */
		.text
		.align  5
LEAF(spurious_interrupt)
		/*
		 * Someone tried to fool us by sending an interrupt but we
		 * couldn't find a cause for it.
		 */
		lui     t1,%hi(irq_err_count)
		lw      t0,%lo(irq_err_count)(t1)
		addiu   t0,1
		sw      t0,%lo(irq_err_count)(t1)
		j	ret_from_irq
END(spurious_interrupt)

		__INIT

		.set	reorder

NESTED(except_vec1_generic, 0, sp)
		PANIC("Exception vector 1 called")
END(except_vec1_generic)

		/*
		 * General exception vector.  Used for all CPUs except R4000
		 * and R4400 SC and MC versions.
		 */
NESTED(except_vec3_generic, 0, sp)
#if defined(R5432_CP0_INTERRUPT_WAR)
		mfc0	k0, CP0_INDEX
#endif
		mfc0	k1, CP0_CAUSE
		la	k0, exception_handlers
		andi	k1, k1, 0x7c
		addu	k0, k0, k1
		lw	k0, (k0)
		jr	k0
END(except_vec3_generic)
		.set	at

		/* General exception vector R4000 version. */
NESTED(except_vec3_r4000, 0, sp)
		.set	push
		.set    mips3
		.set	noat
		mfc0	k1, CP0_CAUSE
		li	k0, 31<<2
		andi	k1, k1, 0x7c
		.set	noreorder
		beq	k1, k0, handle_vced
		 li	k0, 14<<2
		beq	k1, k0, handle_vcei
		 la	k0, exception_handlers
		.set	reorder
		addu	k0, k0, k1
		lw	k0, (k0)
		jr	k0

		/*
		 * Big shit, we now may have two dirty primary cache lines for
		 * the same physical address.  We can savely invalidate the
		 * line pointed to by c0_badvaddr because after return from
		 * this exception handler the load / store will be re-executed.
		 */
handle_vced:
		mfc0	k0, CP0_BADVADDR
		li	k1, -4
		and	k0, k1
		mtc0	zero, CP0_TAGLO
		cache	Index_Store_Tag_D,(k0)
		cache	Hit_Writeback_Inv_SD,(k0)
#ifdef CONFIG_PROC_FS
		lui	k0, %hi(vced_count)
		lw	k1, %lo(vced_count)(k0)
		addiu	k1, 1
		sw	k1, %lo(vced_count)(k0)
#endif
		eret

handle_vcei:
		mfc0	k0, CP0_BADVADDR
		cache	Hit_Writeback_Inv_SD, (k0)	# also cleans pi
#ifdef CONFIG_PROC_FS
		lui	k0, %hi(vcei_count)
		lw	k1, %lo(vcei_count)(k0)
		addiu	k1, 1
		sw	k1, %lo(vcei_count)(k0)
#endif
		eret
		.set    pop
END(except_vec3_r4000)

		__FINIT

/*
 * Build a default exception handler for the exceptions that don't need
 * special handlers.  If you didn't know yet - I *like* playing games with
 * the C preprocessor ...
 */
#define __BUILD_clear_none(exception)
#define __BUILD_clear_sti(exception)                                    \
		STI
#define __BUILD_clear_cli(exception)                                    \
		CLI
#define __BUILD_clear_fpe(exception)                                    \
		cfc1	a1,fcr31;                                       \
		li	a2,~(0x3f<<12);                                 \
		and	a2,a1;                                          \
		ctc1	a2,fcr31;                                       \
		STI
#define __BUILD_clear_ade(exception)                                    \
		.set	reorder;					\
		MFC0	t0,CP0_BADVADDR;                                \
		.set	noreorder;					\
		REG_S	t0,PT_BVADDR(sp);                               \
		KMODE
#define __BUILD_silent(exception)

#define fmt "Got %s at %08lx.\n"

#define __BUILD_verbose(exception)                                      \
		la	a1,8f;                                          \
		TEXT	(#exception);                                   \
		REG_L	a2,PT_EPC(sp);                                  \
		PRINT(fmt)
#define __BUILD_count(exception)                                        \
		.set	reorder;                                        \
		lw	t0,exception_count_##exception;                 \
		.set	noreorder;                                      \
		addiu	t0, 1;                                          \
		sw	t0,exception_count_##exception;                 \
		.data;                                                  \
EXPORT(exception_count_##exception);                                    \
		.word	0;                                              \
		.previous;


#define BUILD_HANDLER(exception,handler,clear,verbose)                  \
		.align	5;                                              \
NESTED(handle_##exception, PT_SIZE, sp);		                \
		.set	noat;                                           \
		SAVE_ALL;                                               \
FEXPORT(handle_##exception##_int);					\
		__BUILD_clear_##clear(exception);                       \
		.set	at;                                             \
		__BUILD_##verbose(exception);                           \
		jal	do_##handler;                                   \
		 move	a0, sp;                                         \
		j	ret_from_exception;                             \
		 nop;                                                   \
END(handle_##exception)

		BUILD_HANDLER(adel,ade,ade,silent)		/* #4  */
		BUILD_HANDLER(ades,ade,ade,silent)		/* #5  */
		BUILD_HANDLER(ibe,be,cli,silent)		/* #6  */
		BUILD_HANDLER(dbe,be,cli,silent)		/* #7  */
		BUILD_HANDLER(bp,bp,sti,silent)			/* #9  */
		BUILD_HANDLER(ri,ri,sti,silent)			/* #10 */
		BUILD_HANDLER(cpu,cpu,sti,silent)		/* #11 */
		BUILD_HANDLER(ov,ov,sti,silent)			/* #12 */

		.align	5 /*#13 */
		NESTED(handle_tivo_tr, PT_SIZE, sp)
		.set	noat
		__BUILD_silent(tr)
		.set	push
		.set	reorder
		.extern	tivo_time_hi,4
		/* Check for fast path traps, if none found goto C-Based exception. */
	
		/* First check for counter trap. 0x50f4 */
		mfc0	k1,CP0_EPC
		lw	k0,0(k1)
		xori	k0, 0x50f4
#ifdef CONFIG_FUNCTION_COUNTER
		
		.set noreorder
		beqz    k0, grab_counter
		 xori	k0, 0x50f4 /* Can't hork k1 here since grab_counter needs it */
	
		/*
		 * Next check for lower order trap function counter
		 * which is tnei which has top 16bits of 0x072e
		 * and the lower bits encoded with the number of
		 * the function.  If true, re-load the instruction
		 * opcode in k0 to pass as an arg to record_fn_num
		 * which uses the bottom 17 bits of k0 to figure out
		 * where in the function map to put stuff.  We can
		 * do this since 0x072E has the LSB set to 0.
		 *
		 */

		srl     k1, k0, 16 
		xori    k1, 0x072e
		beqz    k1, record_fn_num

		/*
		 * Now check for higher order trap function counter
		 * which has upper 16 bits of 0 0x07ae
		 *
		 */
		 xori    k1, 0x0080
		bnez    k1, handle_tr /* Not a special trap, goto dflt logic */

		/*
		 * The function number for this trap (0x07ae) starts at 0x10000,
		 * and record_fn_num expects 17 bits.  So, it's legit to bitwise or
		 * our opcode which contains the bottom 16 with 0x10000.
		 * The top bits will be discarded by record_fn_num.
		 *
		 */

		 lui     k1, 0x1 /* BDS legal, handle_exception doesn't assume k0 */
		or	k0, k1

record_fn_num:
		/*
		 * When we get to record_fn_num, we know we're executing
		 * one of our function counter traps.  So, we have to
		 * emulate the instruction that our trap was inserted
		 * into.  (addu gp, t9)  Move this to the BDS as it's
		 * the only thing we can schedule there.
		 *
		 */

		sll     k0, 15 /* Keeping both I-pipes on 5432 running when possible */

		lw	k1, funcount_buffer
		beqz	k1, 5f
		 nop
		srl     k0, 13 
		addu	k1, k0, k1
		andi	k0, k1, 1 /* Check for Funorder instead of funcount */
		bnez	k0, funorder
		 nop
		lw	k0, 0(k1)
		addiu	k0, 1
		bnezl	k0, 5f
		 sw	k0, 0(k1)
	
5:		mfc0    k1, CP0_EPC /* EPC update must occur *after* touch to
				     * vmalloc'd memory, otherwise we're
				     * going to miss Trap's in the fault case.
				     */
		addi    k1, 4
		mtc0    k1, CP0_EPC
		addu    gp, t9
		nop
		nop
		.set mips3
		eret
		.set mips0
funorder:
		xori	k1, 0x1 /* Clear bottom bit for real address */
		lw	k0, 0(k1)
		bnez	k0, 5b
		lui	k0, %hi(funorder_count)
		lw	k0, %lo(funorder_count)(k0)
		addiu	k0, 1
		sw	k0, 0(k1)
		lui	k1, %hi(funorder_count)
		b	5b
		sw	k0, %lo(funorder_count)(k1)
	
		.set reorder
#else
		bnez    k0, handle_tr
#endif
grab_counter:	addi	k1, 4;
		mtc0	k1, CP0_EPC
		lw 	v0, tivo_time_hi
		mfc0 	v1, CP0_COUNT
		mfc0	k0, CP0_CAUSE
		sll	k0,k0,16
		srl	k0,k0,31
		bnez	k0,2f
		.set mips3
		eret
		.set mips0
2:		mfc0	v1, CP0_COMPARE
		sltiu	v1, v1, 1	/* Only add 1 if CP0_COMPARE is 0 */
		add	v0, v1
		mfc0 	v1, CP0_COUNT
		.set mips3
		eret
		.set mips0
		.set	pop
		.set	at
END(handle_tivo_tr)

		BUILD_HANDLER(tr,tr,sti,silent) 	/* fall through #13 */
		BUILD_HANDLER(fpe,fpe,fpe,silent)	/* #15 */
		BUILD_HANDLER(watch,watch,cli,silent)	/* #23 */
		BUILD_HANDLER(mcheck,mcheck,cli,silent)	/* #24 */
		BUILD_HANDLER(reserved,reserved,sti,silent)	/* others */

		.set	pop

/*
 * Table of syscalls
 */
		.data
		.align	PTRLOG
EXPORT(sys_call_table)
#define SYS(call, narg) PTR call

		/* Reserved space for all SVR4 syscalls. */
		.space	(1000)*PTRSIZE

#ifdef CONFIG_BINFMT_IRIX
		/* 32bit IRIX5 system calls. */
#include "irix5sys.h"
#else
		.space	(1000)*PTRSIZE		/* No IRIX syscalls */
#endif

		/* Reserved space for all the BSD43 and POSIX syscalls. */
		.space	(2000)*PTRSIZE

		/* Linux flavoured syscalls. */
#include "syscalls.h"

/*
 * Number of arguments of each syscall
 */
EXPORT(sys_narg_table)
#undef SYS
#define SYS(call, narg) .byte narg

		/* Reserved space for all SVR4 flavoured syscalls. */
		.space	(1000)

#ifdef CONFIG_BINFMT_IRIX
		/* 32bit IRIX5 system calls. */
#include "irix5sys.h"
#else
		.space	(1000)			/* No IRIX syscalls */
#endif

		/* Reserved space for all the BSD43 and POSIX syscalls. */
		.space	(2000)

		/* Linux flavoured syscalls. */
#include "syscalls.h"
