/*
 *  gen06-pci.c
 *
 *  Based largely on:
 *  arch/mips/brcm/brcmpci.c -- Broadcom BCM93730 pci setup routines
 *
 *  Copyright (C) 2000 Troy Trammel <ttrammel@broadcom.com>
 *					Broadcom Corp.
 *
 *  Modifications Copyright (C) 2006 by TiVo, Inc.
 * 
 */

#include <linux/init.h>
#include <linux/kernel.h>
#define CONFIG_PCI 
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/ioport.h>

#include <asm/bcm7401.h>
#include <asm/io.h>

#define dprintk(...)

#include <asm/pci_channel.h>

#ifdef DUMP_PCI_CONFIG
static void PrintPCIDev( struct pci_dev *pci_dev )
{
    printk(
        "unsigned int   devfn                %08X  SLOT %02X FUNC %02X\n"
        "unsigned short vendor               %04X\n"
        "unsigned short device;              %04X\n"  
        "unsigned short subsystem_vendor     %04X\n" 
        "unsigned short subsystem_device     %04X\n"
        "unsigned int   class                %08X\n"
        "u8             hdr_type             %02X\n"
        "u8             rom_base_reg         %02X\n"
        "struct pci_driver *driver           %08X\n"
        "void           *driver_data         %08X\n"
        "dma_addr_t     dma_mask             %08X\n"
        "char           name[80]             %s\n"      
        "char           slot_name[8]         %s\n"
        "int            active               %08X\n"
        "int            ro                   %08X\n"
        "unsigned short regs                 %08X\n",
        pci_dev->devfn,  PCI_SLOT(pci_dev->devfn), PCI_FUNC(pci_dev->devfn),
        pci_dev->vendor,
        pci_dev->device,
        pci_dev->subsystem_vendor,
        pci_dev->subsystem_device,
        pci_dev->class,
        pci_dev->hdr_type,
        pci_dev->rom_base_reg,
        (int)pci_dev->driver,
        (int)pci_dev->driver_data,
        (int)pci_dev->dma_mask,
        pci_dev->name,
        pci_dev->slot_name,
        pci_dev->active,
        pci_dev->ro,
        pci_dev->regs);
}
#endif

spinlock_t pciconf_lock = SPIN_LOCK_UNLOCKED;
static struct resource pci_io_resource = {
	"io pci IO space",
	0x0000000,
	0x0000000 + BCM_PCIIO_SIZE,
	IORESOURCE_IO,
	NULL, NULL, NULL
};

static struct resource pci_mem_resource = {
	"pci memory space",
	BCM_PCIMEM_PHYS,
	BCM_PCIMEM_PHYS+ BCM_PCIMEM_SIZE,
	IORESOURCE_MEM,
	NULL, NULL, NULL
};

static struct resource pci_sata_io_resource = {
	"io sata pci IO space",
	0,/*SATA_IO_BASE*/
	0 /*SATA_IO_BASE*/ + SATA_IO_SIZE,
	IORESOURCE_IO,
	NULL, NULL, NULL
};

static struct resource pci_sata_mem_resource = {
	"sata pci memory space",
	SATA_MEM_BASE,
	SATA_MEM_BASE+ SATA_MEM_SIZE,
	IORESOURCE_MEM,
	NULL, NULL, NULL
};

static int brcm_pci_read_config_dword( struct pci_dev *dev,
				int where, u32 *val)
{
    int slot_num, func_num;
    u32 cfc,cf8;
    unsigned long flags;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

    switch (dev->bus->number) {
    case 1: 
	/*
	 *On 7401 SATA device in slot 0 is responding to all PCI discovery
	 *queries as if it were connected on all possible slots. (Bug???)
	 */ 
	if (dev->devfn != 0) 
	{
	    return PCIBIOS_FUNC_NOT_SUPPORTED;
	}
	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_DATA; 
    break;
    case 0:
	cfg_index_reg	=   (volatile u32 *)MIPS_PCI_XCFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)MIPS_PCI_XCFG_DATA; 
    break;
    default:

    /*
     *   configuration cycle 0 only (primary bus only)
     */
         printk("brcm_pci_write_config_dword: Invalid bus number\n");
	  return PCIBIOS_FUNC_NOT_SUPPORTED;
    }

    slot_num = PCI_SLOT(dev->devfn);	
    func_num = PCI_FUNC(dev->devfn);
    cf8 = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

    spin_lock_irqsave( &pciconf_lock, flags );

    *cfg_index_reg = cf8;
    cfc = *cfg_data_reg;
    *val = cfc;
    
    spin_unlock_irqrestore( &pciconf_lock, flags );
    
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_dword(struct pci_dev *dev, int where,
					u32 val)
{
    int slot_num, func_num;
    u32 cfc;
    unsigned long flags;
    volatile u32 *cfg_index_reg;
    volatile u32 *cfg_data_reg;

    switch (dev->bus->number) {
    case 1: 
	/*
	 *On 7401 SATA device in slot 0 is responding to all PCI discovery
	 *queries as if it were connected on all possible slots. (Bug???)
	 */ 
	if (dev->devfn != 0) 
	{
	    return PCIBIOS_FUNC_NOT_SUPPORTED;
	}
	cfg_index_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)PCI_BRIDGE_SATA_CFG_DATA; 
    break;		
    case 0:
	cfg_index_reg	=   (volatile u32 *)MIPS_PCI_XCFG_INDEX;
	cfg_data_reg	=   (volatile u32 *)MIPS_PCI_XCFG_DATA; 
    break;
    default:

    /*
     *   configuration cycle 0 only (primary bus only)
     */
       printk("brcm_pci_write_config_dword: Invalid bus number\n");
	return PCIBIOS_FUNC_NOT_SUPPORTED;
    }

    slot_num = PCI_SLOT(dev->devfn);
    func_num = PCI_FUNC(dev->devfn);

    cfc = PCI_ENABLE + PCI_IDSEL(slot_num) + PCI_FNCSEL(func_num) + where;

    spin_lock_irqsave( &pciconf_lock, flags );

    *cfg_index_reg = cfc;
    *cfg_data_reg = val;

    spin_unlock_irqrestore( &pciconf_lock, flags );

    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	result >>= 16;
    *val = result & 0xffff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
    int status;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 1)
	result >>= 8;
    if (where & 2)
	result >>= 16;
    *val = result & 0xff;
    return PCIBIOS_SUCCESSFUL;
}

static int brcm_pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
    int status, shift = 0;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    result &= ~(0xffff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where & ~3, result);
}

static int brcm_pci_write_config_byte( struct pci_dev *dev, int where, u8 val)
{
    int status, shift = 0;
    u32 result;

    status = brcm_pci_read_config_dword(dev, where & ~3, &result);
    if (status != PCIBIOS_SUCCESSFUL)
	return status;
    if (where & 2)
	shift += 16;
    if (where & 1)
	shift += 8;
    result &= ~(0xff << shift);
    result |= val << shift;
    return brcm_pci_write_config_dword(dev, where & ~3, result);
}

struct pci_ops brcm_pci_ops = {
    brcm_pci_read_config_byte,
    brcm_pci_read_config_word,
    brcm_pci_read_config_dword,
    brcm_pci_write_config_byte,
    brcm_pci_write_config_word,
    brcm_pci_write_config_dword
};

struct pci_channel mips_pci_channels[] = {
	/* Primary PCI bus */
	{ &brcm_pci_ops, &pci_io_resource, &pci_mem_resource, 1, 0xff },

	/* Secondary PCI bus only contains 1 SATA controller at devfn 0 */
	{ &brcm_pci_ops, &pci_sata_io_resource, &pci_sata_mem_resource, 0, 0},
	{ NULL, NULL, NULL, 0, 0}
};

static void bcm_pcierr_interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
    printk( KERN_WARNING "PCI Error detected, IRQ=%d @0x%08lx\n", irq, regs->cp0_epc );
}

void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;
	int slot_num;
	int bus_num;

	pci_for_each_dev(dev) {

		slot_num = PCI_SLOT(dev->devfn);
		bus_num = dev->bus->number;
		switch (bus_num) 
		{
		    case 0:
			switch (slot_num) 
			{
			    case PCI_DEVICE_ID_EXT:
				    break;
                case PCI_DEVICE_ID_VIXS:
#ifdef CONFIG_MIPS_BRCM974XX
                 pcibios_update_irq(dev, BCM_IRQ_PCI_INTA_0_CPU_INTR);
                 dev->irq = BCM_IRQ_PCI_INTA_0_CPU_INTR;
#else
                 pcibios_update_irq(dev, BCM_IRQ_PCI_INTA_1_CPU_INTR);
                 dev->irq = BCM_IRQ_PCI_INTA_1_CPU_INTR;
#endif /* CONFIG_MIPS_BRCM974XX*/
                 break;
			    default:
				    printk("unknown slot num 0x%x on bus 0x%x for"
                            "pcibios_fixup_irqs\n", slot_num,bus_num);
			}
		    break;
		    case 1:
			switch (slot_num) 
			{
			    case PCI_DEVICE_ID_SATA:	/* IDE controller */
				    pcibios_update_irq(dev, BCM_IRQ_PCI_SATA_CPU_INTR);
				    dev->irq = BCM_IRQ_PCI_SATA_CPU_INTR;
                break;
			    default:
				    printk("unknown slot num 0x%x on bus 0x%x "
                            "for pcibios_fixup_irqs\n", slot_num,bus_num);
			}
		    break;
		    default:
			    printk("unknown bus num 0x%x for pcibios_fixup_irqs\n", bus_num);
		}
	}

	request_irq( BCM_IRQ_PCI_ADR_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SERR_DET_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_MSTR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_TAR_ABORT_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_REC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_SRC_DATA_PERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_DMA_ERROR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
	request_irq( BCM_IRQ_PCI_RBUS_MSTR_ERR_INTR, &bcm_pcierr_interrupt, 
	             SA_INTERRUPT, "pci error", NULL );
}

// Remembers which devices have already been assigned.
static  int slots[32];

static volatile u32 memAddr = CPU2PCI_CPU_PHYS_MEM_WIN_BASE;
static volatile u32 ioAddr = CPU2PCI_CPU_PHYS_IO_WIN0_BASE;

void account_mapped_mem(struct pci_dev *dev,
                    unsigned int *msize,
                    unsigned int *iosize,
                    int pnp){
	int i,j,k,l;
	u32 savRegData;
	u32 size;
	u32 mask;
	u32 regData;
	u32 tregData;

	/* Write FFFFFFFFH to BAR registers to probe for Mem */
    if ( iosize ) *iosize = 0;
    if ( msize) *msize = 0;
	for (i=0,j=0; i<6; i++,j+=4) {
		pci_read_config_dword(dev,PCI_BASE_ADDRESS_0+j,&tregData);
		pci_write_config_dword(dev,PCI_BASE_ADDRESS_0+j,0xffffffff);
		pci_read_config_dword(dev,PCI_BASE_ADDRESS_0+j,&regData);
		pci_write_config_dword(dev,PCI_BASE_ADDRESS_0+j,tregData);
		if (regData) {
			dprintk("PCI PnP: PCI_BAR[%d] = %x\n", i, regData);

			savRegData = regData;
            /* Get rid of Bit 0-3 0=Mem, 2:1 32/64, 3=Prefetchable */
			regData &= 0xfffffff0; 
			/*
			* determine io/mem size here by looking at the BAR
			* register, skipping the first 4 bits
			*/
			for (k=4; k<32; k++) {
				if (regData & (1<<k)) {
					break;
				}
			}
			if (k < 32) {
				dprintk("PCI PnP: size requested is %x\n", 1<<k);
				size = 1<<k;
				mask = 0xffffffff;
				for (l=0; l<k; l++) {
					mask &= ~(1<<l);
				}
			}
			else
				break;
			if (savRegData & 0x1) {	/* IO address */
                    /* Calculate the next address that satisfies the boundary condition */
                    regData = (ioAddr + size - 1) & mask;
                if ( pnp ){
                    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, regData);
                    dprintk("PCI PnP: Writing PCI_IO_BAR[%d]=%x, size=%d, mask=%x\n", i,
                        regData, 1<<k, mask);
                }
                if ( iosize )
				    *iosize = regData + size;
				ioAddr = regData + size; // Advance it
			}
			else { 
				/* Calculate the next address that satisfies the boundary condition */
				regData = (memAddr + size - 1) & ~(size -1);
                if ( pnp ){
                    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0+j, regData);
                    dprintk("PCI PnP: Writing PCI_MEM_BAR[%d]=%x, size=%d, mask=%x\n", i,
                        regData, 1<<k, mask);
                }
                if ( msize )
                    *msize = regData + size;
				memAddr = regData +size; // Advance it
			}

		}
	}
}

/* Do IO/memory allocation for expansion slot here */
static __init void brcm_pcibios_fixup_plugnplay(struct pci_dev *dev) 
{
	int slot;
	u32 regData;
    unsigned int iosize=0, msize=0;
		
	/*
	* Make sure that it is in the PCI slot, otherwise we punt.
	*/

	slot = PCI_SLOT(dev->devfn);

	if (slots[slot]) {
		printk("Skip resource assignment for dev ID "
                "%04x:%04x already allocated\n", dev->vendor, dev->device);
		return;
	}

	slots[slot] = 1;

	pcibios_update_irq(dev, BCM_IRQ_PCI_INTA_2_CPU_INTR);
	dev->irq = BCM_IRQ_PCI_INTA_2_CPU_INTR;

	printk("\tPCI DEV in slot %x, ID=%04x:%04x\n\r", slot, dev->vendor, dev->device);
    
    account_mapped_mem(dev,&msize,&iosize,1);

	/*  now that it's all configured, turn on the Command Register */
	pci_read_config_dword(dev,PCI_COMMAND,&regData);
	regData |= PCI_COMMAND_SERR|PCI_COMMAND_PARITY |PCI_COMMAND_MASTER;
	if (iosize > 0)
		regData |= PCI_COMMAND_IO;
	if (msize > 0)
		regData |= PCI_COMMAND_MEMORY;
	pci_write_config_dword(dev,PCI_COMMAND,regData);
	printk("\tPCI PnP: PCI_DEV %04x:%04x, command=%x, memAddr=%08x, ioAddr=%08x\n", 
		dev->vendor, dev->device, regData, memAddr, ioAddr);
	
}
void __init pcibios_update_irq(struct pci_dev *dev, int irq)
{
    u8 val;

	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	/* Test this as it does not seem to be working */
	pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &val);
}
unsigned __init int pcibios_assign_all_busses(void)
{
	return 0;
}
void __init pcibios_fixup_resources(struct pci_dev *dev)
{
    return;
}
void __init brcm_pcibios_fixup_device(struct pci_dev *dev)
{
	u32 dev_vendor_id;
	u32 regData;
	u32 size;

	/********************************************************
	 * Configure the Device
	 ********************************************************/
	dev_vendor_id = (u32)dev->device<<16;
	dev_vendor_id |= dev->vendor;

	switch (dev_vendor_id) {
	    /********************************************************
	     * SATA Controller
	     ********************************************************/
	    case BCM7401_SATA_VID:
	    {
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_0, SATA_PCS0_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_1, SATA_PCS1_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_2, SATA_SCS0_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_3, SATA_SCS1_OFS+1);
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_4, SATA_CFG_BUS_MASTER_OFS+1);

            account_mapped_mem(dev,NULL,NULL,0);
		    
		    size = CPU2PCI_PCI_SATA_PHYS_MEM_WIN0_SIZE;
		    regData = CPU2PCI_PCI_SATA_PHYS_MEM_WIN0_BASE; 
		    pci_write_config_dword(dev, PCI_BASE_ADDRESS_5, regData);

		    regData = 0x0f;
            /* Minimum Grant, to avoid Latency being reset to lower value */
		    pci_write_config_byte(dev, 0x3e, regData); 
		    
		    regData = 0xff;
            /* SATA_CFG_MASTER_LATENCY_TIMER. */
		    pci_write_config_byte(dev, 0xd, regData); 

		    pci_read_config_dword(dev, PCI_COMMAND, &regData);
		    regData |= (PCI_COMMAND_SERR|PCI_COMMAND_PARITY|
                        PCI_COMMAND_MASTER|PCI_COMMAND_IO| PCI_COMMAND_MEMORY);
		    pci_write_config_dword(dev, PCI_COMMAND, regData);

		    /* Primary */
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);

		    regData &= 0xfffffc; /* Mask off  Reserved & Res Type bits.*/

		    regData += 2; // Offset 302H for primary
		    outb(inb(regData)|0x60,regData);

            /* Secondary */
		    pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);

		    regData &= 0xfffffc; // Mask off  Reserved & Res Type bits.

            regData += 0xa; // Offset 30AH for secondary
            outb(inb(regData)|0x60,regData);

#ifdef CONFIG_TIVO_SATAPM
			if ( dev->bus->number == 1 ){
            	extern void EnableAndShowPmSignature();
            	EnableAndShowPmSignature();
			}
#endif // CONFIG_TIVO_SATAPM
	    }
	    break;

		case TIVO_VIXS_PCI_ID:
		{
            pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,memAddr);
            pci_write_config_dword(dev,PCI_BASE_ADDRESS_1,memAddr+0x100000);

            account_mapped_mem(dev,NULL,NULL,0);

            pci_read_config_dword(dev,PCI_COMMAND,&regData);
            regData |= PCI_COMMAND_SERR|PCI_COMMAND_PARITY |PCI_COMMAND_MASTER;
            regData |= PCI_COMMAND_MEMORY;
            pci_write_config_dword(dev,PCI_COMMAND,regData);

#if DEBUG_PCI
            u8 byteVal;
            pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &regData);
            printk(" ViXS BAR0 0x%x\n", regData);

            pci_write_config_dword(dev,PCI_BASE_ADDRESS_1,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &regData);
            printk(" ViXS BAR1 0x%x\n", regData);

            pci_write_config_dword(dev,PCI_BASE_ADDRESS_2,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &regData);
            printk(" ViXS BAR2 0x%x\n", regData);

            pci_write_config_dword(dev,PCI_BASE_ADDRESS_3,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_3, &regData);
            printk(" ViXS BAR3 0x%x\n", regData);

            pci_write_config_dword(dev,PCI_BASE_ADDRESS_4,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_4, &regData);
            printk(" ViXS BAR4 0x%x\n", regData);

            pci_write_config_dword(dev,PCI_BASE_ADDRESS_5,0xffffffff);
            pci_read_config_dword(dev, PCI_BASE_ADDRESS_5, &regData);
            printk(" ViXS BAR5 0x%x\n", regData);

            pci_read_config_byte(dev, PCI_INTERRUPT_LINE, &byteVal);
            printk(" ViXS PCI interrupt line = 0x%x\n", byteVal);

            pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &byteVal);
            printk(" ViXS PCI interrupt pin = 0x%x\n", byteVal);
#endif /*DEBUG_PCI*/

		}
		break;

        default:
		    printk("unknown dev_man id 0x%x, trying expansion slot\n", dev_vendor_id);
			{
		    	/* For 7401, it is not really Expansion slot any more, 
				 * we just let the Plug&Play logic does the assignment */
		    	brcm_pcibios_fixup_plugnplay(dev);
			}

	    break;
	}
#ifdef DUMP_PCI_CONFIG
    PrintPCIDev( dev );
	{
	    int i;
	    printk("\tPCI Config Space IDSEL%d",PCI_SLOT(dev->devfn));
	    for (i=0;i<0x40;i++) {
		    if ((i & 3) == 0) printk("\r\n\t\t");
		    pci_read_config_byte(dev,i,(unsigned char *)&regData);
		    printk("[%.2x]=%.2x  ",i,regData & 0xFF);
	    }
	    printk("\n\r");
	}
#endif
	return;
}
static void add_pci_wired_tlbs(void){
    unsigned long lo0, lo1, hi;
    int entry,entries;
    unsigned int memSize = (memAddr-CPU2PCI_CPU_PHYS_MEM_WIN_BASE);

    hi = (CPU2PCI_CPU_PHYS_MEM_WIN_BASE & 0xffffe000);
    lo0 = (((CPU2PCI_CPU_PHYS_MEM_WIN_BASE>>(4+2))&0x3fffffc0)|0x17);
    lo1 = ((((CPU2PCI_CPU_PHYS_MEM_WIN_BASE+OFFSET_16MBYTES)>>(4+2))&0x3fffffc0)|0x17);

    entry = 0;

    entries = (memSize+ (OFFSET_16MBYTES - (memSize & (OFFSET_16MBYTES -1))))/
                OFFSET_16MBYTES;
    // Wire a 1:1 physical address mapping for this address range.
    // Use 16M page mappings, the largest we can get.
    // XXX: needs review.
    while (entry < entries) {
    add_wired_entry(lo0, lo1, hi, PM_16M);
            hi += OFFSET_32MBYTES;
            lo0 += TLBLO_OFFSET_32MBYTES;
            lo1 += TLBLO_OFFSET_32MBYTES;
            entry++;
    }
}

void __init pcibios_fixup(void)
{
    struct pci_dev *dev;
    pci_for_each_dev(dev) {
		/* <TDT> Our enable_device sets up bios data for the bootrom does not */
		brcm_pcibios_fixup_device(dev);

		/* <TDT> force a rescan in case we still need bios data */
		pci_setup_device(dev);
    }
    dprintk("Total PCI mem needed is %x\n",memAddr-CPU2PCI_CPU_PHYS_MEM_WIN_BASE);
    dprintk("Total PCI io needed is %x\n",ioAddr-CPU2PCI_CPU_PHYS_IO_WIN0_BASE);
    add_pci_wired_tlbs();
}


void pcibios_set_master( struct pci_dev *dev )
{
    u16 command;
    unsigned long flags;
 
    save_and_cli(flags);
    pci_read_config_word(dev, PCI_COMMAND, &command);
    command |= PCI_COMMAND_MASTER;
    pci_write_config_word(dev, PCI_COMMAND, command);
    restore_flags(flags);
}
