/*
 *  Reset routines - for now, just restart the firmware at reset vector
 *
 * Copyright (C) 2000 TiVo Inc. All Rights Reserved.
 */

#include <linux/config.h>
#include <linux/mm.h>
#include <linux/sysrq.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>

void (*tivo_board_reboot)(void);

void tivo_restart(char *command)
{
	void *reset_vec = (void*)0xBFC00000;

	flush_cache_all();

        turnoff_caches();
        clear_caches(); /*paranoid*/

	if (tivo_board_reboot)
		tivo_board_reboot();
	goto *reset_vec;
}

void tivo_halt(void)
{
	void *reset_vec = (void*)0xBFC00000;

        flush_cache_all();
        
        turnoff_caches();
        clear_caches(); /*paranoid*/
    
	if (tivo_board_reboot)
		tivo_board_reboot();
	goto *reset_vec;
}

void tivo_power_off(void)
{
	void *reset_vec = (void*)0xBFC00000;

        flush_cache_all();
        
        turnoff_caches();
        clear_caches(); /*paranoid*/
    
	if (tivo_board_reboot)
		tivo_board_reboot();
	goto *reset_vec;
}

// Following was swiped wholly from sysrq.c:
//
static int is_local_disk(kdev_t dev)	    /* Guess if the device is a local hard drive */
{
	unsigned int major = MAJOR(dev);

	switch (major) {
	case IDE0_MAJOR:
	case IDE1_MAJOR:
	case IDE2_MAJOR:
	case IDE3_MAJOR:
		return 1;
	default:
		return 0;
	}
}

static void go_sync(struct super_block *sb)
{
	printk(KERN_INFO "Syncing device %s ... ",
	       kdevname(sb->s_dev));

	fsync_dev(sb->s_dev);			    /* Sync only */
	printk("OK\n");
}

volatile int emergency_sync_scheduled;

void do_emergency_sync(void)
{
	struct super_block *sb;

	lock_kernel();
	emergency_sync_scheduled = 0;

	for (sb = sb_entry(super_blocks.next);
	     sb != sb_entry(&super_blocks); 
	     sb = sb_entry(sb->s_list.next))
		if (is_local_disk(sb->s_dev))
			go_sync(sb);

	for (sb = sb_entry(super_blocks.next);
	     sb != sb_entry(&super_blocks); 
	     sb = sb_entry(sb->s_list.next))
		if (!is_local_disk(sb->s_dev) && MAJOR(sb->s_dev))
			go_sync(sb);

	unlock_kernel();
	printk(KERN_INFO "Done.\n");
}
