/*
   Quick and dirty raw dump of the pre-processor database.
 */

/* Copyright, compile requirements.

   (c) Keith Owens <kaos@ocs.com.au> January 2002.
   Released under GPL version 2, (you know the rest).
 */

#define NEED_PPDB_INTERNALS
#include "pp_db.h"
#include "pp_makefile.h"
#include <fcntl.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

const char *program = "pp_dbdump";

int errors;
int norun;
int verbose;
int signalled;

PPDB *db;

static const DB_TOKEN db_null_token = DB_TOKEN_NULL;

static int
open_db(void)
{
	fprintf(stderr, "Dumping %s\n", ppdb_name);
	db = open_db_main(2);
	return (0);
}

static int
check_length_eq(datum d, int l)
{
	if (d.dsize != l) {
		printf("\n*** wrong length data, should be %d, is %d\n", l,
		       d.dsize);
		return (0);
	}
	return (1);
}

static int
check_length_ge(datum d, int l)
{
	if (d.dsize < l) {
		printf("\n*** wrong length data, should be >= %d, is %d\n", l,
		       d.dsize);
		return (0);
	}
	return (1);
}

static void inline
print_flag(int flag, const char *text)
{
	if (flag) {
		printf("  %s", text);
	}
}

static void
print_db_token(const char *id, DB_TOKEN p)
{
	if (memcmp(&p, &db_null_token, sizeof (p))) {
		if (id) {
			printf("  %s", id);
		}
		printf(" %c-%d", p.type, p.recno);
		if (p.type == DB_RECTYPE_TEXT) {
			datum key, value;
			printf(" (");
			key.dptr = &p;
			key.dsize = sizeof (p);
			value = ppdb_fetch(db, key);
			if (value.dptr) {
				printf("'%s'", (char *) (value.dptr));
			} else {
				printf("?" "??");
			}
			printf(")");
		}

	}
}

static void
print_db_token_list(const char *id, DB_TOKEN p)
{
	if (memcmp(&p, &db_null_token, sizeof (p))) {
		if (id) {
			printf("  %s", id);
		}
		printf(" %c-%d", p.type, p.recno);
	}
}

static void
dump_base_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_BASE_LIST))) {
		DB_REC_BASE_LIST *r = d.dptr;
		int i;
		printf("base\n");
		printf("  count %d", r->count);
		printf("\n");
		if (!check_length_eq(d, DB_SIZE_BASE_LIST(r))) {
			return;
		}
		for (i = 0; i < r->count; ++i) {
			printf("  base[%d]", i);
			if (r->list[i].flags.has_modules) {
				printf("  has_modules");
			}
			if (r->list[i].flags.has_arch_head) {
				printf("  has_arch_head");
			}
			if (r->list[i].flags.is_host) {
				printf("  is_host");
			}
			print_db_token("name", r->list[i].entry);
			print_db_token("topdir", r->list[i].topdir);
			printf("\n");
		}
	}
}

static void
dump_filename_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_FILENAME_LIST))) {
		DB_REC_FILENAME_LIST *r = d.dptr;
		DB_REC_FILENAME_1 *s;
		int i;
		printf("filename_list\n");
		printf("  count %d", r->count);
		printf("\n");
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			printf("  filename[%d]  index %3d", i, s->index);
			print_db_token("entry", s->entry);
			printf("\n");
		}
	}
}

static void
dump_directory(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_DIRECTORY))) {
		DB_REC_DIRECTORY *r = d.dptr;
		printf("directory\n");
		print_db_token("relname", r->relname);
		print_db_token("parent", r->parent);
		print_flag(r->flags.create_objdir, "create_objdir");
		print_flag(r->flags.explicit, "explicit");
		print_flag(r->flags.walked, "walked");
		print_flag(r->flags.makefile, "makefile");
		print_db_token("extra_aflags", r->extra_aflags);
		print_db_token("extra_cflags", r->extra_cflags);
		print_db_token("extra_ldflags", r->extra_ldflags);
		print_db_token("strip_aflags", r->strip_aflags);
		print_db_token("strip_cflags", r->strip_cflags);
		print_db_token("strip_ldflags", r->strip_ldflags);
		print_db_token_list("children", r->children);
		printf("\n");
	}
}

static void
dump_base_select_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_BASE_SELECT_LIST))) {
		DB_REC_BASE_SELECT_LIST *r = d.dptr;
		DB_REC_BASE_SELECT_1 *s;
		int i;
		printf("base_select_list\n");
		printf("  count %d\n", r->count);
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			if (!DB_IS_NULL_TOKEN(s->entry)) {
				printf("  select[%d]", i);
				print_db_token("entry", s->entry);
			}
			printf("\n");
		}
	}
}

static void
dump_select_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_SELECT_LIST))) {
		DB_REC_SELECT_LIST *r = d.dptr;
		DB_REC_SELECT_1 *s;
		int i;
		printf("select_list\n");
		printf("  count %d\n", r->count);
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			printf("  select[%d]", i);
			print_flag(s->flags.sely, "sely");
			print_flag(s->flags.selm, "selm");
			print_flag(s->flags.sels, "sels");
			print_flag(s->flags.objlinked, "objlinked");
			print_flag(s->flags.walked, "walked");
			print_db_token("entry", s->entry);
			printf("\n");
		}
	}
}

static void
dump_target(datum d)
{
	if (check_length_eq(d, sizeof (DB_REC_TARGET))) {
		DB_REC_TARGET *r = (DB_REC_TARGET *) (d.dptr);
		printf("target\n");
		print_db_token("relname", r->relname);
		print_db_token("parent", r->parent);
		printf("  mtime %ld", r->mtime);
		print_flag(r->flags.nodepend, "nodepend");
		print_flag(r->flags.setup, "setup");
		print_flag(r->flags.expsyms, "expsyms");
		print_flag(r->flags.archive, "archive");
		print_flag(r->flags.archive_member, "archive_member");
		print_flag(r->flags.object, "object");
		print_flag(r->flags.assembler, "assembler");
		print_flag(r->flags.gen_i, "gen_i");
		print_flag(r->flags.gen_s, "gen_s");
		print_flag(r->flags.makecmdgoal, "makecmdgoal");
		print_flag(r->flags.makefile, "makefile");
		print_flag(r->flags.commands, "commands");
		print_flag(r->flags.commands_gen, "commands_gen");
		print_flag(r->flags.arch_head, "arch_head");
		print_flag(r->flags.sely_summary, "sely_summary");
		print_flag(r->flags.selm_summary, "selm_summary");
		print_flag(r->flags.host_summary, "host_summary");
		print_flag(r->flags.issrcfile, "issrcfile");
		print_flag(r->flags.uptodate, "uptodate");
		print_flag(r->flags.multi_linked, "multi_linked");
		print_flag(r->flags.warned_multi_linked, "warned_multi_linked");
		print_flag(r->flags.walked, "walked");
		print_flag(r->flags.ignore, "ignore");
		print_flag(r->flags.symlink, "symlink");
		print_db_token_list("source", r->source);
		print_db_token_list("dependency", r->dependency);
		print_db_token_list("objlink", r->objlink);
		print_db_token("command", r->command);
		print_db_token("extra_aflags", r->extra_aflags);
		print_db_token("extra_cflags", r->extra_cflags);
		print_db_token("extra_ldflags", r->extra_ldflags);
		print_db_token("strip_aflags", r->strip_aflags);
		print_db_token("strip_cflags", r->strip_cflags);
		print_db_token("strip_ldflags", r->strip_ldflags);
		print_db_token("link", r->link);
		printf("\n");
	}
}

static void
dump_target_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_TARGET_LIST))) {
		DB_REC_TARGET_LIST *r = d.dptr;
		DB_REC_TARGET_1 *s;
		int i;
		printf("target_list\n");
		printf("  count %d", r->count);
		printf("\n");
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			printf("  target[%d]", i);
			print_db_token("entry", s->entry);
			printf("\n");
		}
	}
}

static void
dump_source(datum d)
{
	if (check_length_eq(d, sizeof (DB_REC_SOURCE))) {
		DB_REC_SOURCE *r = (DB_REC_SOURCE *) (d.dptr);
		printf("source\n");
		print_db_token("relname", r->relname);
		print_db_token("parent", r->parent);
		printf("  index %3d", r->index);
		printf("  mtime %ld", r->mtime);
		print_flag(r->flags.uptodate, "uptodate");
		print_flag(r->flags.config_extracted, "config_extracted");
		print_flag(r->flags.shadow_source, "shadow_source");
		print_flag(r->flags.setup, "setup");
		print_db_token_list("component", r->component);
		print_db_token_list("new_component", r->new_component);
		print_db_token_list("config", r->config);
		printf("\n");
	}
}

static void
dump_child_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_CHILD_LIST))) {
		DB_REC_CHILD_LIST *r = d.dptr;
		DB_REC_CHILD_1 *s;
		int i;
		printf("child_list\n");
		printf("  count %d", r->count);
		print_flag(r->flags.sorted, "sorted");
		printf("\n");
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			printf("  child[%d]", i);
			print_flag(!s->flags.primary, "       ");
			print_flag(s->flags.primary, "primary");
			printf("  index %3d", s->index);
			print_db_token("relname", s->relname);
			print_db_token("entry", s->entry);
			printf("\n");
		}
	}
}

static void
dump_ccd_list(datum d)
{
	if (check_length_ge(d, sizeof (DB_REC_CCD_LIST))) {
		DB_REC_CCD_LIST *r = d.dptr;
		DB_REC_CCD_1 *s;
		int i;
		printf("ccd_list\n");
		printf("  count %d", r->count);
		print_flag(r->flags.sorted, "sorted");
		printf("  parent_index %3d", r->parent_index);
		printf("  parent_mtime %ld", r->parent_mtime);
		printf("\n");
		for (i = 0; i < r->count; ++i) {
			s = r->list + i;
			printf("  ccd[%d]", i);
			print_db_token("entry", s->entry);
			printf("  index %3d", s->index);
			printf("  mtime %ld", s->mtime);
			printf("\n");
		}
	}
}

static void
dump_text(datum d)
{
	printf("text %d", d.dsize);
	printf("\n");
}

struct save_token {
	DB_TOKEN tok;
	const char *filename;
	datum val;
};

static int
qsort_tokens(const void *va, const void *vb)
{
	const struct save_token *a = va, *b = vb;
	if (a->tok.type < b->tok.type) {
		return (-1);
	}
	if (a->tok.type > b->tok.type) {
		return (1);
	}
	if (a->tok.type == DB_RECTYPE_FILENAME_LIST) {
		return (strcmp(a->filename, b->filename));
	}
	if (a->tok.recno < b->tok.recno) {
		return (-1);
	}
	if (a->tok.recno > b->tok.recno) {
		return (1);
	}
	return (0);
}

static void
dump_db_header(void)
{
	printf("Header\n");
	printf("  version %d fixed_slab_num %d fixed_key_vector_size %d\n"
	       "  variable_shift_min %d variable_shift_max %d hashsize %d\n",
	       db->header->version, db->header->fixed_slab_num,
	       db->header->fixed_key_vector_size,
	       db->header->variable_shift_min, db->header->variable_shift_max,
	       db->header->hashsize);
	printf("  header %p self %p size %d dirty %d next_recno %d\n",
	       db->header, db->header->self, db->header->size,
	       db->header->flags.dirty, db->header->next_recno);
	printf("  source_slab %p free %d\n",
	       db->header->source_slab, db->header->source_slab_free);
	printf("  target_slab %p free %d\n",
	       db->header->target_slab, db->header->target_slab_free);
	printf("  directory_slab %p free %d\n",
	       db->header->directory_slab, db->header->directory_slab_free);
	printf("  free_start %p free_left %d\n",
	       db->header->free_start, db->header->free_left);
}

static int
dump_db(void)
{
	kvpair p;
	DB_TOKEN tok;
	const static kvpair nullkv;
	struct save_token *sk = NULL;
	int i, save_token_alloc = 0, save_token_used = 0;
	dump_db_header();
	for (p = ppdb_first(db); memcmp(&p, &nullkv, sizeof (p));
	     p = ppdb_next(db)) {
		memcpy(&tok, (void *) (p.key.dptr), sizeof (tok));
		if (save_token_used == save_token_alloc) {
			sk = realloc(sk,
				     (save_token_alloc += 2000) * sizeof (*sk));
		}
		sk[save_token_used].val = p.val;
		sk[save_token_used].tok = tok;
		if (tok.type == DB_RECTYPE_FILENAME_LIST) {
			sk[save_token_used++].filename = p.key.dptr;
		} else {
			sk[save_token_used].filename = NULL;
			if (p.key.dsize != sizeof (tok)) {
				printf("%c *** wrong size %d ***\n",
				       tok.type, p.key.dsize);
			} else {
				++save_token_used;
			}
		}
	}
	qsort(sk, save_token_used, sizeof (*sk), qsort_tokens);
	for (i = 0; i < save_token_used; ++i) {
		tok = sk[i].tok;
		printf("token");
		if (tok.type == DB_RECTYPE_FILENAME_LIST) {
			printf(" %s  ", sk[i].filename);
		} else {
			print_db_token(NULL, tok);
			printf("  ");
		}
		switch (tok.type) {
		case DB_RECTYPE_BASE_LIST:
			dump_base_list(sk[i].val);
			break;
		case DB_RECTYPE_FILENAME_LIST:
			dump_filename_list(sk[i].val);
			break;
		case DB_RECTYPE_DIRECTORY:
			dump_directory(sk[i].val);
			break;
		case DB_RECTYPE_BASE_SELECT_LIST:
			dump_base_select_list(sk[i].val);
			break;
		case DB_RECTYPE_SELECT_LIST:
			dump_select_list(sk[i].val);
			break;
		case DB_RECTYPE_TARGET:
			dump_target(sk[i].val);
			break;
		case DB_RECTYPE_TARGET_LIST:
			dump_target_list(sk[i].val);
			break;
		case DB_RECTYPE_SOURCE:
			dump_source(sk[i].val);
			break;
		case DB_RECTYPE_CHILD_LIST:
			dump_child_list(sk[i].val);
			break;
		case DB_RECTYPE_CCD_LIST:
			dump_ccd_list(sk[i].val);
			break;
		case DB_RECTYPE_TEXT:
			dump_text(sk[i].val);
			break;
		default:
			fprintf(stderr,
				"unknown record, type %c (%d), recno %d, length %d\n",
				tok.type, tok.type, tok.recno, sk[i].val.dsize);
		}
	}
	return (0);
}

static int
close_db(void)
{
	close_db_main(db);
	return (0);
}

int
main(int argc, char **argv)
{
	int ret;
	if (argc != 1) {
		fprintf(stderr, "%s takes no parameters\n", program);
		exit(1);
	}
	/* Dummy call to pp_getenv() for srctree defines.  Never executed. */
	if (argc == 2) {
		pp_getenv();
	}
	ret = open_db() || dump_db() || close_db();
	return (ret);
}
